
from cython cimport boundscheck, wraparound, cdivision
from cython.view cimport contiguous

import numpy as np


ctypedef fused jf_data_t:
    unsigned short  # raw pixel data
    float  # corrected pixel data
    unsigned int  # mask data
    unsigned char  # gain data


DEF STRIXEL_Y = 86
DEF STRIXEL_X = 1024 * 3 + 18


def strixel_shape():
    return STRIXEL_Y, STRIXEL_X


def strixel_double_pixels():
    """Build index arrays for double-size pixels.

    In raw data, the entire columns 255, 256, 511, 512, 767 and 768
    are double-size pixels. After strixelation, these end up in columns
    765-776, 1539-1550 and 2313-2324 on rows 0-85 or 0-83, with a set
    of four columns with 86 rows followed by a set of 84 and 86 again.

    This function builds the index arrays after strixelation.
    """

    Ydouble = []
    Xdouble = []

    for double_col in [765, 1539, 2313]:
        for col in range(double_col, double_col+12):
            for row in range(84 if ((col-double_col) // 4) == 1 else 86):
                Ydouble.append(row)
                Xdouble.append(col)

    return np.array(Ydouble), np.array(Xdouble)


@boundscheck(False)
@wraparound(False)
@cdivision(True)
def strixel_transform(jf_data_t[:, :, ::contiguous] data,
                      jf_data_t[:, :, ::contiguous] out = None):
    """Reorder raw data to physical strixel sensor layout. """

    if data.shape[1] < 256 or data.shape[2] < 256:
        raise ValueError('Pixel shape of data may not be below (256, 256)')

    if out is None:
        import numpy as np
        out = np.zeros((data.shape[0], STRIXEL_Y, STRIXEL_X), dtype=np.float32)
    elif data.shape[0] > out.shape[0]:
        raise ValueError('Cell shape of data exceeds out')
    elif out.shape[1] < STRIXEL_Y or out.shape[2] < STRIXEL_X:
        raise ValueError(f'Pixel shape of out may not be below '
                         f'({STRIXEL_Y}, {STRIXEL_X})')

    cdef int cell, yin, xin, xout, yout, igap

    for cell in range(data.shape[0]):
        # Normal pixels.
        for yin in range(256):
            yout = yin // 3

            for xin in range(1024) :
                xout = 774 * (xin // 256) + 3 * (xin % 256) + yin % 3
                out[cell, yout, xout] = data[cell, yin, xin]

        # Gap pixels.
        for yin in range(256):
            yout = 2 * (yin // 6)

            for igap in range(3) :
                # Left side of the gap.
                xin = igap * 256 + 255
                xout = igap * 774 + 765 + yin % 6
                out[cell, yout, xout] = data[cell, yin, xin]
                out[cell, yout+1, xout] = data[cell, yin, xin]

                # Right side of the gap.
                xin = igap * 256 + 255 + 1
                xout = igap * 774 + 765 + 11 - yin % 6
                out[cell, yout, xout] = data[cell, yin, xin]
                out[cell, yout+1, xout] = data[cell, yin, xin]

    return out
